using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using InABox.Core;
using InABox.WPF;
using Syncfusion.UI.Xaml.Grid;

namespace InABox.DynamicGrid;

public class DynamicGridCheckBoxColumn<TEntity> : DynamicGridEditorColumn<TEntity,CheckBoxEditor,GridTemplateColumn>
    where TEntity : BaseObject
{
    private readonly BitmapImage ticked = Wpf.Resources.Bullet_Tick.AsBitmapImage();
    private readonly BitmapImage unticked = null; //Wpf.Resources.tick.AsGrayScale().AsBitmapImage();
    
    protected override void Configure(GridTemplateColumn column, CheckBoxEditor editor)
    {
        column.CellTemplate = TemplateGenerator.CreateDataTemplate
        (
            () =>
            {
                var result = new Image() { Margin = new Thickness(2) };
                result.SetBinding(
                    Image.SourceProperty, 
                    new Binding()
                    {
                        Path = new PropertyPath(MappingName), 
                        Converter = new BoolToImageConverter()
                    }
                );
                return result;
            }
        );
        
        column.EditTemplate = TemplateGenerator.CreateDataTemplate
        (
            () =>
            {
                DockPanel dock = new DockPanel();
                dock.Children.Add(CreateImage(ticked, Visibility.Visible, Visibility.Collapsed, BindingMode.TwoWay));
                dock.Children.Add(CreateImage(unticked, Visibility.Collapsed, Visibility.Visible, BindingMode.TwoWay));
                Border border = new Border()
                {
                    Background = new SolidColorBrush(Colors.White),
                    Child = dock
                };
                
                border.PreviewMouseDown += (sender, args) =>
                {
                    if (sender is Border b && b.Child is DockPanel d)
                        ToggleImages(d);
                    args.Handled = true;
                };
                border.KeyUp += (sender, args) =>
                {
                    if (args.Key == Key.Space && sender is Border b && b.Child is DockPanel d)
                        ToggleImages(d);
                    args.Handled = true;
                };
                return border;
            }
            
        );
        column.SetCellBoundValue = false;
    }

    private static void ToggleImages(DockPanel d)
    {
        var t = d.Children.OfType<Image>().First();
        var u = d.Children.OfType<Image>().Last();
        t.Visibility = t.Visibility == Visibility.Visible
            ? Visibility.Collapsed
            : Visibility.Visible;
        u.Visibility = u.Visibility == Visibility.Visible
            ? Visibility.Collapsed
            : Visibility.Visible;
    }

    private Image CreateImage(BitmapImage source, Visibility truevalue, Visibility falsevalue, BindingMode mode)
    {
        var image = new Image() { Source = source, Margin = new Thickness(2) };
        image.SetValue(DockPanel.DockProperty,Dock.Top);
        image.SetBinding(
            Image.VisibilityProperty, 
            new Binding()
            {
                Path = new PropertyPath(MappingName), 
                Converter = new BoolToVisibilityConverter() { TrueValue = truevalue, FalseValue = falsevalue }, 
                Mode = mode
            }
        );
        image.PreviewMouseDown += (sender, args) =>
        {
            if (sender is Image i)
                i.Visibility = i.Visibility == truevalue
                    ? falsevalue
                    : truevalue;
            args.Handled = true;
        };
        image.KeyUp += (sender, args) =>
        {
            if (args.Key == Key.Space && sender is Image i)
                i.Visibility = i.Visibility == truevalue
                    ? falsevalue
                    : truevalue;
            args.Handled = true;
        };
        return image;
    }

    public override bool Filtered => false;
    public override bool VariableWidth => false;

    public DynamicGridCheckBoxColumn(DynamicGridColumn definition) : base(definition)
    {
    }
}